


library(R.utils)
library(reshape2)
library(mice)
library(igraph)
library(class)
library(randomForest)
library(statnet)
library(Bergm)



## adt = mlits[[1]]
makedf<- function(dat){
  adjmat <- dat[["adj_matrix"]]
  ind <- apply(adjmat, 2, sum, na.rm=T)
  outd <- apply(adjmat, 1, sum, na.rm=T)
  

  # add data for imputation, omit covariate c to divorce from simulation
  adjlist <- melt(adjmat) #var1 is outgoing actor, var2 is incoming actor
  adjlist$ind <- ind[adjlist$Var1]
  adjlist$outd <- outd[adjlist$Var1]
  adjlist$homophilyA <- dat[["va"]][adjlist$Var1] - dat[["va"]][adjlist$Var2]
  adjlist$homophilyB <- dat[["vb"]][adjlist$Var1] - dat[["vb"]][adjlist$Var2]
  lbdfunc <- function(rowdat){
    tmp<-ifelse(adjmat[rowdat["Var2"], rowdat["Var1"]]==1,1,0)
    if (is.na(tmp)){
      return(0)
    }
    return(tmp)
  }
  adjlist$recip<-apply(adjlist, 1, lbdfunc)
  return (adjlist)
}



list2mat<-function(adjlist){
  imputed_matrix<-dcast(adjlist, Var1~Var2)
  rownames(imputed_matrix) <-imputed_matrix$Var1
  imputed_matrix<-imputed_matrix[,-1]
  colnames(imputed_matrix)<-rownames(imputed_matrix) 
  return(imputed_matrix)
}

impute_knn <- function(adjlist){
  train <- adjlist[complete.cases(adjlist),]
  dens<-sum(train$value)/nrow(train)
  test <-  adjlist[!complete.cases(adjlist),]
  pred <- knn(train=train[,4:ncol(train)], test=test[,4:ncol(train)], cl=train[,3], k=3, prob=T)
  probs<-attributes(pred)$prob
  probs[pred==0]<-(1-probs[pred==0])
  predv<-ifelse(probs>dens,1,0)
  adjlist[!complete.cases(adjlist),"value"]<- predv
  return (list2mat(adjlist[1:3]))
}

impute_mice <-function(adjlist){
  pred <- mice(adjlist[,3:ncol(adjlist)],m=30)
  mats <- list()
  for (i in 1:30){
    imp <- pred$imp$value[,i]
    adjlist_imp <- adjlist
    adjlist_imp[!complete.cases(adjlist),]$value<-imp
    mats[[i]] <- list2mat(adjlist_imp)
  }
  return (mats)
}

impute_RF <- function (adjlist){
  train <- adjlist[complete.cases(adjlist),]
  test <-  adjlist[!complete.cases(adjlist),]
  dens<-sum(train$value)/nrow(train)
  train[,3]<-factor(train[,3])
  rf <- randomForest( value ~ .,  data=train[,3:ncol(adjlist)], ntree = 100)
  pred <- predict(rf, newdata=test[, 4:ncol(adjlist)], cutoff=c(1-dens, dens))
  pred <- as.numeric(levels(pred)[pred])
  adjlist[!complete.cases(adjlist),]$value<-pred
  return (list2mat(adjlist[1:3]))
}

impute_reconstruction <- function(networkdat){
  # impute tie^i,j with tie^j,i
  transposed <- t(networkdat)
  missindex <- which(is.na(networkdat))
  networkdat[missindex] <- transposed[missindex]
  
  # find density of graph, impute tie^i,j where i j both are missing actors
  g <- graph_from_adjacency_matrix(networkdat)
  density <- edge_density(g)
  missindex <- which(is.na(networkdat)) # now only misssing ties between missing actors
  #set.seed(2048)
  networkdat[missindex]<-ifelse(runif(length(missindex))<density, 1, 0)
  return(networkdat)
}


impute_CRDPG <- function(networkdat, n){
  # create a graph base on dataset with NAs
  
  networkdat<-as.matrix(networkdat)
  networkdat.copy <- networkdat
  networkdat.copy[which(is.na(networkdat.copy))] <- 0
  
  missindex<-which(is.na(networkdat))
  
  tryCatch({
    graph <- graph_from_adjacency_matrix(networkdat.copy)
    
    # create latent embedding of graph
    embed <- embed_adjacency_matrix(graph, n)
    
    for (ind in missindex){
      
      i = floor(ind/dim(networkdat)[1])+1
      j = ind %% dim(networkdat)[1]
      
      outvec<-embed$X[i,]
      invec<-embed$Y[j,]
      prob<-sum(outvec*invec)
      if(prob>1){
        networkdat[ind]<-1
      }else if (prob<0){
        networkdat[ind]<-0
      }else{
        if (runif(1)<prob){
          networkdat[ind]<-1
        }else{
          networkdat[ind]<-0
        }
      }
      
    }
    
  },error = function(e){NULL})
  
  if(sum(is.na(networkdat))>0){
    networkdat<-matrix(NA,nrow=100, ncol=100)
  }
  
  return(as.matrix(networkdat))
}

impute_PA <- function(networkdat){
  networkdat<-as.matrix(networkdat)
  # get outdegree distribution
  NAs<-rowSums(is.na(networkdat))
  rsum <- rowSums(networkdat, na.rm=T)
  rsum <- rsum[NAs<(nrow(networkdat)-1)]
  
  # get Ji all set of actors no tie to target
  missactors <- networkdat[rowSums(is.na(networkdat)) > 0,] #rows ith missing outgoing ties
  #set.seed(2048)
  rsum_sample <- 0
  if (length(rsum)>=nrow(missactors)){
    rsum_sample <- rsum[sample(length(rsum),nrow(missactors),replace = F)] # sample nrow random outdegrees
  }else{
    rsum_sample <- rsum[sample(length(rsum),nrow(missactors),replace = T)] 
  }
  
  
  # outdegree difference (num ties to draw)
  outdegree_diff <- rsum_sample-rowSums(missactors, na.rm = T)
  #print(outdegree_diff)
  outdegree_diff[outdegree_diff<0] <-0
  #print(outdegree_diff)
  
  # PA probability
  indegrees <- colSums(networkdat, na.rm = T)
  PApmf <- sort(indegrees/sum(indegrees))
  PApdf <- Reduce("sum", PApmf, accumulate = TRUE)
  names(PApdf) <- names(PApmf)
  
  networkdat.PA <- data.frame(networkdat)
  
  for(i in 1:nrow(missactors)){
    intvec <- PApdf[-which(names(PApdf) %in% names(outdegree_diff[i]))] #don't consider the imputed actor itself
    intvec2 <- intvec[-which(names(intvec) %in% names(missactors[i,][!is.na(missactors)]))] # dont consider actors already have ties with the target
    indtie <- findInterval(runif(outdegree_diff[i]), intvec2) 
    ties <- names(PApdf[indtie]) 
    networkdat.PA[rowSums(is.na(networkdat)) > 0,][i,] <- 0
    networkdat.PA[rowSums(is.na(networkdat)) > 0,][i,unique(ties)] <- 1
    
    
  }
  
  networkdat.PA[!is.na(networkdat)]<- networkdat[!is.na(networkdat)]
  return(as.matrix(networkdat.PA))
}


impute_null <- function(networkdat){
  # create a graph base on dataset with NAs
  networkdat<-as.matrix(networkdat)
  networkdat[is.na(networkdat)]<-0
  return(as.matrix(networkdat))
}


impute_bergm_simp<-function(netdata, va, vb, vc, n){
  net <- as.network(netdata)
  set.vertex.attribute(net, "va", va)
  set.vertex.attribute(net, "vb", vb)
  set.vertex.attribute(net, "vc", vc)
  
  imp<-NULL
  withTimeout( 
    {tryCatch({
      imp <- bergmM(net ~ edges + mutual+ nodecov("va") + nodematch("vb"),
                    burn.in = 1000,
                    aux.iters = 2000,
                    main.iters = 3000,
                    gamma = 0.5,
                    prior.sigma = diag(4, 4),
                    nImp = n)
      
    },error = function(e){NULL})},
    timeout = 3600,
    onTimeout = "silent")
  

  impnets<-list()
  if (length(imp$impNets)<n){
    for (x in 1:n){
      impnets[[x]]<-netdata
    }
  }else{
    impnets <- lapply(imp$impNets, as.matrix)
  }
  # print(length(imp$impNets))
  # print(length(impnets))
  
  return(impnets)
  
}

impute_bergm_complex<-function(netdata, va, vb, vc, n){
  net <- as.network(netdata)
  set.vertex.attribute(net, "va", va)
  set.vertex.attribute(net, "vb", vb)
  set.vertex.attribute(net, "vc", vc)
  
  imp<-NULL
  withTimeout(
    {tryCatch({
      imp <- bergmM(net ~ edges + mutual +  gwesp(0.5, fixed=T) + nodematch("vb") + nodematch("vc") + nodecov("va") + nodefactor("vb") + nodecov("vc"),
                    burn.in = 1000,
                    aux.iters = 2000,
                    main.iters = 3000,
                    gamma = 0.5,
                    prior.sigma = diag(4,8),
                    nImp = n)

    },error = function(e){NULL})},
    timeout = 3600,
    onTimeout = "silent")




  
 
  impnets<-list()
  if (length(imp$impNets)<n){
    for (x in 1:n){
      impnets[[x]]<-netdata
    }
  }else{
    impnets <- lapply(imp$impNets, as.matrix)
  }
  
  return(impnets)
  
}




# 
# adjlist<-makedf(mlist[[1]])
# imp_knn<-impute_knn(adjlist)
# imp_mice<-impute_mice(adjlist)
# imp_rf<- impute_RF(adjlist)
# imp_re<-impute_reconstruction(mlist[[1]]$adj_matrix)
# imp_pa<-impute_PA(mlist[[1]]$adj_matrix)
# imp_dp<-impute_CRDPG(mlist[[1]]$adj_matrix,5)
# imp_null<-impute_null(mlist[[1]]$adj_matrix)
# imp_bergm_simp<-impute_bergm_simp(mlist[[1]]$adj_matrix, mlist[[1]]$va, mlist[[1]]$vb, mlist[[1]]$vc, 5)
# imp_bergm_comp<-impute_bergm_complex(mlist[[1]]$adj_matrix, mlist[[1]]$va, mlist[[1]]$vb, mlist[[1]]$vc, 5)
# 
# View(imp_null)
# View(mlist[[1]]$adj_matrix)


