library(ergm)
library(abind)




# simulate complete data based on ERGM
# arguments: n is sample size, coefs is the coefficient list for network statistics, nsim is the number of simulated datasets

ergm_sim <- function (n, coefs, nsim = 2){
  
  net <- network(n) # empty network
  va <- rnorm(n, mean=20, sd=3)  # covariate a
  vb <- ifelse(runif(n)>0.4, 1, 2)   # covariate b
  # covariate c 0.1, 0.3, 0.6, 0.85, 1
  randnum <- runif(n)
  vc <- ifelse (randnum < 0.1, 1, ifelse(randnum<0.3, 2, ifelse(randnum<0.6, 3, ifelse(randnum<0.85,4,5))))
  set.vertex.attribute(net, "va", va)
  set.vertex.attribute(net, "vb", vb)
  set.vertex.attribute(net, "vc", vc)
  
  # ERGM
  sim1 <- simulate(net ~ edges + mutual + gwesp(0.5, fixed=T) + nodematch("vb") + nodematch("vc") + nodecov("va") + nodefactor("vb") + nodecov("vc"), 
                   coef = coefs,
                   nsim = nsim,
                  control=control.simulate.formula(MCMC.burnin=10000, MCMC.interval=200000))
  
  
  #print(summary(sim1~edges + mutual + idegree1.5 + odegree1.5 + nodecov("va") + nodematch("vb") + nodematch("vc")))
  
  sim1_matrices<-lapply(sim1, as.matrix)
  sim1_matrices <- lapply(sim1_matrices, function(x) {colnames(x) <- paste0("v",c(1:n));x})
  sim1_matrices <- lapply(sim1_matrices, function(x) {rownames(x) <- paste0("v",c(1:n));x})
  names(va) <- paste0("v",c(1:n))
  names(vb) <- paste0("v",c(1:n))
  names(vc)<- paste0("v",c(1:n))
  result_all<-list()
  for (i in 1:nsim){
    result<-list()
    result[["adj_matrix"]]<-sim1_matrices[[i]]
    result[["va"]]<-va
    result[["vb"]]<-vb
    result[["vc"]]<-vc
    result_all[[i]]<-result
  }
  
  return(result_all)
  
}


##### functions for missing data generation

# mcar actor missing
# n is num actors, p is missing percent, dat is adjacency matrix
mcar_actor_missing <- function(n, p, dat){
  # print(n)
  # print(p)
  nums <- sample(nrow(dat), p*n)
  dat.mcar <- dat
  dat.mcar[nums, ] <- NA
  diag(dat.mcar)<-0
  return(dat.mcar)
}


# mar actor missing based on b. 80% of missing value from 1, 20% from 2
# account for cases when the num that should be missing exceeds category 1/2, then remove the rest of the missing values from the other category
mar_actor_missing <- function(n, p, dat, v){
  mmiss <- 0
  
  if(length(v[v==1])>=0.8*p*n){
    mmiss <- names(v[v==1][sample(length(v[v==1]), 0.8*p*n)])
  }else{
    mmiss <- names(v[v==1])
  }
  fmiss<-0
  if(length(v[v==2])>=p*n-length(mmiss)){
    fmiss <- names(v[v==2][sample(length(v[v==2]), p*n-length(mmiss))])
  }else{
    fmiss_1 <- names(v[v==2])
    rest<-v[v==1][which(!v[v==1] %in% mmiss)]
    fmiss_2 <- rest[sample(length(rest),p*n-length(mmiss)-length(fmiss_1))]
    fmiss<-c(fmiss_1, fmiss_2)
  }
  
  
  
  mindex <- match(mmiss, colnames(dat))
  findex <- match(fmiss, colnames(dat))
  dat.mar <- dat
  dat.mar[mindex, ] <- NA
  dat.mar[findex, ] <- NA
  diag(dat.mar)<-0
  return(dat.mar)
}

# mnar actor based on outdegree
mnar_actor_missing1 <- function(n, p, dat){
  miss <- names(sort(rowSums((dat)))[1:(p*n)])
  index <- match(miss, colnames(dat))
  dat.mnar <- dat
  dat.mnar[index, ] <- NA
  diag(dat.mnar)<-0
  return(dat.mnar)
}

# mnar actor based on indegree
mnar_actor_missing2 <- function(n, p, dat){
  miss <- names(sort(colSums((dat)))[1:(p*n)])
  index <- match(miss, colnames(dat))
  dat.mnar <- dat
  dat.mnar[index, ] <- NA
  diag(dat.mnar)<-0
  return(dat.mnar)
}


mcar_cell_missing <- function(n, p, dat){
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  missind <- sample(allind_noiden,floor(p*n*(n-1)))
  dat.mcar <- dat
  dat.mcar[missind] <- NA
  diag(dat.mcar)<-0
  return(dat.mcar)
}

# mar cell missing based on b. 80% of missing value from 1, 20% from 2
# account for cases when the num that should be missing exceeds category 1/2, then remove the rest of the missing values from the other category
mar_cell_missing <- function(n, p, dat, v){
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  rows <-  (allind_noiden %% n)
  cols <- floor( allind_noiden/n)+1
  rows[rows==0] <- 50
  # mar probability based on row attribute v
  v_rows <- v[rownames(dat[rows,])]
  index1 <- allind_noiden[v_rows==1]
  index2 <- allind_noiden[v_rows==2]
  
  missindex1<-0
  if(length(index1)>= 0.8*p*n*(n-1)){
    missindex1 <- index1[sample(length(index1), 0.8*p*n*(n-1))]
  }else{
    missindex1 <- index1
  }
  missindex2<-0
  if(length(index2)>=p*n-length(missindex1)){
    missindex2 <- index2[sample(length(index2), p*n*(n-1)-length(missindex1))]
  }else{
    missindex2_1 <- index2
    rest<-index1[which(!index1 %in% missindex1)]
    missindex2_2 <- rest[sample(length(rest),p*n*length(missindex1)-length(missindex2_1))]
    missindex2<-c(missindex2_1, missindex2_2)
  }
  
  dat.mar <- dat
  dat.mar[missindex1] <- NA
  dat.mar[missindex2] <- NA
  diag(dat.mar)<-0
  return(dat.mar)
}

# more 1's missing (in terms of percentage)
mnar_cell_missing1 <- function(n, p, dat){
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  index1 <- allind_noiden[dat[allind_noiden]==1]
  index2 <- allind_noiden[dat[allind_noiden]==0]
  
  nummiss1<-min(floor(p*n*(n-1)*0.55), floor(length(index1)*0.6))
  nummiss2<-min(p*n*(n-1)-nummiss1, floor(length(index2)*0.8))
  if (nummiss2<p*n*(n-1)-nummiss1){
    nummiss1 = p*n*(n-1)-nummiss2
  }
  
  missindex1<-sample(index1, nummiss1)
  missindex2<-sample(index2, nummiss2)
  
  dat.mnar <- dat
  dat.mnar[missindex1] <- NA
  dat.mnar[missindex2] <- NA
  diag(dat.mnar)<-0
  return(dat.mnar)
}

# higher percentage of 0's missing
mnar_cell_missing2 <- function(n, p, dat){
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  index1 <- allind_noiden[dat[allind_noiden]==1]
  index2 <- allind_noiden[dat[allind_noiden]==0]
  
  
  nummiss1<-min(floor(p*n*(n-1)*0.1), floor(length(index1)*0.3))
  nummiss2<-min(p*n*(n-1)-nummiss1, floor(length(index2)*0.8))
  if (nummiss2<p*n*(n-1)-nummiss1){
    nummiss1 = p*n*(n-1)-nummiss2
  }
  
  missindex1<-sample(index1, nummiss1)
  missindex2<-sample(index2, nummiss2)

  dat.mnar <- dat
  dat.mnar[missindex1] <- NA
  dat.mnar[missindex2] <- NA
  diag(dat.mnar)<-0
  return(dat.mnar)
}



# p is overall missing percent, mixp is percent of missing that are entire actor/rows
mcar_mix_missing <- function(n, p, mixp, dat){
  p1<-floor(mixp*p*n)
  q1<-(p*n-p1)*(n-1)
  nums <- sample(nrow(dat), p1)
  dat.mcar <- dat
  dat.mcar[nums, ] <- NA
  
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  row_miss <- which(is.na(dat.mcar))
  allind_rest <- allind_noiden[which(!allind_noiden %in% row_miss)] # exclude row/actor missing
  missind <- sample(allind_rest,q1)
  dat.mcar[missind] <- NA
  diag(dat.mcar)<-0
  return(dat.mcar)
}


mar_mix_missing <- function(n, p, mixp, dat, v){
  mmiss <- 0
  if(length(v[v==1])>=0.8*mixp*p*n){
    mmiss <- names(v[v==1][sample(length(v[v==1]), 0.8*p*mixp*n)])
  }else{
    mmiss <- names(v[v==1])
  }
  fmiss<-0
  if(length(v[v==2])>=mixp*p*n-length(mmiss)){
    fmiss <- names(v[v==2][sample(length(v[v==2]), floor(mixp*p*n-length(mmiss)))])
  }else{
    fmiss_1 <- names(v[v==2])
    rest<-v[v==1][which(!v[v==1] %in% mmiss)]
    fmiss_2 <- rest[sample(length(rest),floor(mixp*p*n-length(mmiss)-length(fmiss_1)))]
    fmiss<-c(fmiss_1, fmiss_2)
  }
  
  
  mindex <- match(mmiss, colnames(dat))
  findex <- match(fmiss, colnames(dat))
  used<-length(c(mindex, findex))*(n-1)
  
  
  dat.mar <- dat
  dat.mar[mindex, ] <- NA
  dat.mar[findex, ] <- NA
  
  
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  row_miss <- which(is.na(dat.mar))
  allind_rest <- allind_noiden[which(!allind_noiden %in% row_miss)] # exclude row/actor missing
  rows <-  (allind_rest %% n)
  cols <- floor( allind_rest/n)+1
  rows[rows==0] <- 50
  # mar probability based on row attribute v
  v_rows <- v[rownames(dat[rows,])]
  index1 <- allind_rest[v_rows==1]
  index2 <- allind_rest[v_rows==2]
  missindex1<-0
  if(length(index1)>= 0.8*(p*n*(n-1)-used)){
    missindex1 <- index1[sample(length(index1), 0.8*(p*n*(n-1)-used))]
  }else{
    missindex1 <- index1
  }
  
  missindex2<-0
  if(length(index2)>=p*n*(n-1)-used-length(missindex1)){
    missindex2 <- index2[sample(length(index2), p*n*(n-1)-used-length(missindex1))]
  }else{
    missindex2_1 <- index2
    rest<-index1[which(!index1 %in% missindex1)]
    missindex2_2 <- rest[sample(length(rest),p*n*(n-1)-used-length(missindex1)-length(missindex2_1))]
    missindex2<-c(missindex2_1, missindex2_2)
  }
  
  
  dat.mar[missindex1] <- NA
  dat.mar[missindex2] <- NA
  diag(dat.mar)<-0
  return(dat.mar)
}


mnar_mix_missing1 <- function(n, p, mixp, dat){
  p1<-floor(mixp*p*n)
  q1<-(p*n-p1)*(n-1)
  miss <- names(sort(rowSums((dat)))[1:p1])
  index <- match(miss, colnames(dat))
  dat.mnar <- dat
  dat.mnar[index, ] <- NA
  
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  row_miss <- which(is.na(dat.mnar))
  allind_rest <- allind_noiden[which(!allind_noiden %in% row_miss)] # exclude row/actor missing
  
  index1 <- allind_rest[dat[allind_rest]==1]
  index2 <- allind_rest[dat[allind_rest]==0]
  
  
  nummiss1<-min(floor(q1*0.55), floor(length(index1)*0.6))
  nummiss2<-min(q1-nummiss1, floor(length(index2)*0.8))
  if (nummiss2<q1-nummiss1){
    nummiss1 = q1-nummiss2
  }
  
  missindex1<-sample(index1, nummiss1)
  missindex2<-sample(index2, nummiss2)
  
  dat.mnar[missindex1] <- NA
  dat.mnar[missindex2] <- NA
  diag(dat.mnar)<-0
  return(dat.mnar)
}

# row missing indegree, cell 10% 1 missing
mnar_mix_missing2 <- function(n, p, mixp, dat){
  
  p1<-floor(mixp*p*n)
  q1<-(p*n-p1)*(n-1)
  miss <- names(sort(colSums((dat)))[1:p1])
  index <- match(miss, colnames(dat))
  dat.mnar <- dat
  dat.mnar[index, ] <- NA
  
  allind <- c(1:(n*n))
  identity_ind <- c(1:n)
  identities <- (identity_ind-1)*n + identity_ind 
  allind_noiden <- allind[which(!allind %in% identities)] # exclude self-links
  row_miss <- which(is.na(dat.mnar))
  allind_rest <- allind_noiden[which(!allind_noiden %in% row_miss)] # exclude row/actor missing
  
  index1 <- allind_rest[dat[allind_rest]==1]
  index2 <- allind_rest[dat[allind_rest]==0]
  
  
  nummiss1<-min(floor(q1*0.1), floor(length(index1)*0.3))
  nummiss2<-min(q1-nummiss1, floor(length(index2)*0.8))
  if (nummiss2<q1-nummiss1){
    nummiss1 = q1-nummiss2
  }
  
  missindex1<-sample(index1, nummiss1)
  missindex2<-sample(index2, nummiss2)
  
  dat.mnar[missindex1] <- NA
  dat.mnar[missindex2] <- NA
  diag(dat.mnar)<-0
  return(dat.mnar)
}





